import math

import numpy as np
import torch


def uniform_sample_t(num_timesteps, batch_size):
    """Generate a uniform sample of timesteps.
    minimum timestep is 1, maximum timestep is num_timesteps.
    """
    ts = np.random.choice(np.arange(1, num_timesteps + 1), batch_size)
    return torch.from_numpy(ts)


def get_named_beta_schedule(
    schedule_name,
    num_diffusion_timesteps,
    beta_start=0.0001,
    beta_end=0.02,
):
    """Get a pre-defined beta schedule for the given name.

    The beta schedule library consists of beta schedules which remain similar
    in the limit of num_diffusion_timesteps.
    Beta schedules may be added, but should not be removed or changed once
    they are committed to maintain backwards compatibility.
    """
    if schedule_name == "linear":
        return torch.linspace(beta_start, beta_end, num_diffusion_timesteps)
    elif schedule_name == "quadratic":
        return (
            torch.linspace(beta_start**0.5, beta_end**0.5, num_diffusion_timesteps) ** 2
        )
    elif schedule_name == "cosine":
        return cosine_beta_schedule(num_diffusion_timesteps)
    elif schedule_name == "reciprocal":
        return reciprocal_beta_schedule(num_diffusion_timesteps)
    else:
        raise NotImplementedError(f"unknown beta schedule: {schedule_name}")


def cosine_beta_schedule(timesteps, s=0.008):
    """Cosine schedule as proposed in https://arxiv.org/abs/2102.09672.
    """
    steps = timesteps + 1
    x = torch.linspace(0, timesteps, steps)
    alphas_cumprod = torch.cos(((x / timesteps) + s) / (1 + s) * math.pi * 0.5) ** 2
    alphas_cumprod = alphas_cumprod / alphas_cumprod[0]
    betas = 1 - (alphas_cumprod[1:] / alphas_cumprod[:-1])
    return torch.clip(betas, 0.0001, 0.9999)


def reciprocal_beta_schedule(timesteps):
    t = torch.arange(1, timesteps + 1)
    beta_recip = 1 / (timesteps - t + 1)
    return beta_recip



import numpy as np
import torch
import torch.nn.functional as F
from torch import Tensor

from chemeleon_dng.diffusion.models.base import DiffusionModelBase


class DSM(DiffusionModelBase):
    """Denoising Score Matching (DSM).
    """

    def __init__(
        self,
        *,
        num_timesteps: int,
        sigma_begin: float = 0.005,
        sigma_end: float = 0.5,
    ):
        self.timesteps = num_timesteps
        self.sigma_begin = sigma_begin
        self.sigma_end = sigma_end
        _sigmas = torch.FloatTensor(
            np.exp(np.linspace(np.log(sigma_begin), np.log(sigma_end), num_timesteps))
        )
        _sigmas_norm = get_sigma_norm(_sigmas)

        self.sigmas = torch.cat([torch.zeros([1]), _sigmas], dim=0)
        self.sigmas_norm = torch.cat([torch.ones([1]), _sigmas_norm], dim=0)

    def q_sample(self, x_start: Tensor, t: Tensor, batch_idx: Tensor, noise=None):
        """Sample from the forward process q(x_t | x_0).

        :param x_start: the initial data batch.
        :param t: the number of diffusion steps (minus 1). Here, 0 means one step.
        :param batch_idx: the batch index.
        :param noise: if specified, the split-out normal noise.
        """
        if noise is None:
            noise = torch.randn_like(x_start)
        assert noise.shape == x_start.shape
        sigmas = self.sigmas.to(t.device)
        sigma_t = sigmas[t]
        sigma_t_per_atom = sigma_t[batch_idx][:, None]  # [B_n, 1]
        return (x_start + sigma_t_per_atom * noise) % 1.0

    def p_sample(
        self,
        model_output: Tensor,
        x_t: Tensor,
        t: Tensor,
        batch_idx: Tensor,
        noise: Tensor | None = None,
        mode: str = "predictor",
        step_lr: float = 1e-5,
    ):
        sigmas = self.sigmas.to(t.device)
        sigmas_norm = self.sigmas_norm.to(t.device)
        sigma_t = sigmas[t]
        sigma_norm_t = sigmas_norm[t]
        if noise is None:
            noise = torch.randn_like(x_t)
        nonzero_mask = (
            (t[batch_idx] != 1).float().unsqueeze(1).expand(-1, x_t.shape[1])
        )  # no noise when t == 1
        assert nonzero_mask.shape == x_t.shape
        if mode == "predictor":
            adjacent_sigma = sigmas[t - 1]
            step_size = sigma_t**2 - adjacent_sigma**2
            std = torch.sqrt(
                (adjacent_sigma**2 * (sigma_t**2 - adjacent_sigma**2)) / (sigma_t**2)
            )
            # broadcast to [B_n, 1]
            step_size = step_size[batch_idx][:, None]  # [B_n, 1]
            std = std[batch_idx][:, None]  # [B_n, 1]
            sigma_norm_t = sigma_norm_t[batch_idx][:, None]  # [B_n, 1]

            pred_sample = (
                x_t
                - step_size * model_output * torch.sqrt(sigma_norm_t)
                + std * noise * nonzero_mask
            )
            return pred_sample % 1.0  # x_t_minus_05
        elif mode == "corrector":  # x_t = x_t_minus_05
            step_size = step_lr * (sigma_t / self.sigma_begin) ** 2
            std = torch.sqrt(2 * step_size)
            # Broadcast to [B_n, 1]
            step_size = step_size[batch_idx][:, None]  # [B_n, 1]
            std = std[batch_idx][:, None]  # [B_n, 1]
            sigma_norm_t = sigma_norm_t[batch_idx][:, None]  # [B_n, 1]

            pred_sample = (
                x_t
                - step_size * model_output * torch.sqrt(sigma_norm_t)
                + std * noise * nonzero_mask
            )
            return pred_sample % 1.0
        else:
            raise ValueError(f"Unknown mode: {mode}")

    def training_loss(
        self,
        model_output: Tensor,
        noise: Tensor,
        t: Tensor,
        batch_idx: Tensor,
    ):
        sigmas = self.sigmas.to(t.device)
        sigmas_norm = self.sigmas_norm.to(t.device)
        sigma_t = sigmas[t]
        sigma_norm_t = sigmas_norm[t]
        sigma_t_per_atom = sigma_t[batch_idx][:, None]  # [B_n, 1]
        sigma_norm_t_per_atom = sigma_norm_t[batch_idx][:, None]  # [B_n, 1]
        target = d_log_p_wrapped_normal(
            sigma_t_per_atom * noise, sigma_t_per_atom
        ) / torch.sqrt(sigma_norm_t_per_atom)
        loss = F.mse_loss(model_output, target)
        return {"loss": loss}

    def p_mean_and_var(
        self, model_output: Tensor, x_t: Tensor, t: Tensor, batch_idx: Tensor
    ):
        sigmas = self.sigmas.to(t.device)
        sigmas_norm = self.sigmas_norm.to(t.device)
        sigma_t = sigmas[t]
        sigma_norm_t = sigmas_norm[t]
        adjacent_sigma = sigmas[t - 1]

        step_size = sigma_t**2 - adjacent_sigma**2

        # broadcast to [B_n, 1]
        step_size = step_size[batch_idx][:, None]  # [B_n, 1]
        sigma_norm_t = sigma_norm_t[batch_idx][:, None]  # [B_n, 1]

        var = (adjacent_sigma**2 * (sigma_t**2 - adjacent_sigma**2)) / (sigma_t**2)
        var = var[batch_idx][:, None]  # [B_n, 1]
        mean = x_t - step_size * model_output * torch.sqrt(sigma_norm_t)
        return mean, var


def p_wrapped_normal(x, sigma, n=10, t=1.0):
    p_ = 0
    for i in range(-n, n + 1):
        p_ += torch.exp(-((x + t * i) ** 2) / 2 / sigma**2)
    return p_


def d_log_p_wrapped_normal(x, sigma, n=10, t=1.0):
    p_ = 0
    for i in range(-n, n + 1):
        p_ += (x + t * i) / sigma**2 * torch.exp(-((x + t * i) ** 2) / 2 / sigma**2)
    return p_ / p_wrapped_normal(x, sigma, n, t)


def get_sigma_norm(sigma, t=1.0, sn=10000):
    sigmas = sigma[None, :].repeat(sn, 1)
    x_sample = sigma * torch.randn_like(sigmas)
    x_sample = x_sample % t
    normal_ = d_log_p_wrapped_normal(x_sample, sigmas, t=t)
    return (normal_**2).mean(dim=0)  # type: ignore



import tarfile
from pathlib import Path

import requests
from tqdm import tqdm

# Figshare download URL for the checkpoint tar.gz file
FIGSHARE_URL = "https://figshare.com/ndownloader/files/54966305"


def download_file(url: str, filepath: Path) -> None:
    """Download a file from URL with progress bar."""
    response = requests.get(url, stream=True, timeout=30)
    response.raise_for_status()

    total_size = int(response.headers.get("content-length", 0))

    with open(filepath, "wb") as f, tqdm(
        desc=f"Downloading {filepath.name}",
        total=total_size,
        unit="B",
        unit_scale=True,
        unit_divisor=1024,
    ) as pbar:
        for chunk in response.iter_content(chunk_size=8192):
            if chunk:
                f.write(chunk)
                pbar.update(len(chunk))


def extract_tar_gz(filepath: Path, extract_to: Path) -> None:
    """Extract tar.gz file to specified directory."""
    with tarfile.open(filepath, "r:gz") as tar:
        tar.extractall(path=extract_to)


def ensure_checkpoints_downloaded(ckpt_dir: str = "ckpts") -> None:
    """Ensure checkpoints are downloaded and extracted."""
    ckpt_path = Path(ckpt_dir)

    # Check if checkpoint directory exists and has files
    if ckpt_path.exists() and any(ckpt_path.glob("*.ckpt")):
        print("Checkpoints already exist.")
        return

    print("Checkpoints not found. Downloading...")

    # Create checkpoint directory
    ckpt_path.mkdir(parents=True, exist_ok=True)

    # Download tar.gz file
    tar_file = ckpt_path / "checkpoints.tar.gz"
    download_file(FIGSHARE_URL, tar_file)

    # Extract tar.gz file
    print("Extracting checkpoints...")
    extract_tar_gz(tar_file, ckpt_path.parent)

    # Clean up tar.gz file
    tar_file.unlink()

    print(f"Checkpoints downloaded and extracted to {ckpt_path}")


def get_checkpoint_path(task: str, default_paths: dict) -> str:
    """Get checkpoint path, downloading if necessary."""
    ensure_checkpoints_downloaded()

    ckpt_path = default_paths.get(task)
    if ckpt_path and not Path(ckpt_path).exists():
        raise FileNotFoundError(f"Checkpoint not found: {ckpt_path}")

    return ckpt_path



"""basic scatter_sum operations from torch_scatter from
https://github.com/mir-group/pytorch_runstats/blob/main/torch_runstats/scatter_sum.py
Using code from https://github.com/rusty1s/pytorch_scatter, but cut down to avoid a dependency.
PyTorch plans to move these features into the main repo, but until then,
to make installation simpler, we need this pure python set of wrappers
that don't require installing PyTorch C++ extensions.
See https://github.com/pytorch/pytorch/issues/63780.
"""


import torch


def _broadcast(src: torch.Tensor, other: torch.Tensor, dim: int):
    if dim < 0:
        dim = other.dim() + dim
    if src.dim() == 1:
        for _ in range(0, dim):
            src = src.unsqueeze(0)
    for _ in range(src.dim(), other.dim()):
        src = src.unsqueeze(-1)
    src = src.expand_as(other)
    return src


def scatter_sum(
    src: torch.Tensor,
    index: torch.Tensor,
    dim: int = -1,
    out: torch.Tensor | None = None,
    dim_size: int | None = None,
    reduce: str = "sum",
) -> torch.Tensor:
    assert reduce == "sum"
    index = _broadcast(index, src, dim)
    if out is None:
        size = list(src.size())
        if dim_size is not None:
            size[dim] = dim_size
        elif index.numel() == 0:
            size[dim] = 0
        else:
            size[dim] = int(index.max()) + 1
        out = torch.zeros(size, dtype=src.dtype, device=src.device)
        return out.scatter_add_(dim, index, src)
    else:
        return out.scatter_add_(dim, index, src)


def scatter_std(
    src: torch.Tensor,
    index: torch.Tensor,
    dim: int = -1,
    out: torch.Tensor | None = None,
    dim_size: int | None = None,
    unbiased: bool = True,
) -> torch.Tensor:
    if out is not None:
        dim_size = out.size(dim)

    if dim < 0:
        dim = src.dim() + dim

    count_dim = dim
    if index.dim() <= dim:
        count_dim = index.dim() - 1

    ones = torch.ones(index.size(), dtype=src.dtype, device=src.device)
    count = scatter_sum(ones, index, count_dim, dim_size=dim_size)

    index = _broadcast(index, src, dim)
    tmp = scatter_sum(src, index, dim, dim_size=dim_size)
    count = _broadcast(count, tmp, dim).clamp(1)
    mean = tmp.div(count)

    var = src - mean.gather(dim, index)
    var = var * var
    out = scatter_sum(var, index, dim, out, dim_size)

    if unbiased:
        count = count.sub(1).clamp_(1)
    out = out.div(count + 1e-6).sqrt()

    return out


def scatter_mean(
    src: torch.Tensor,
    index: torch.Tensor,
    dim: int = -1,
    out: torch.Tensor | None = None,
    dim_size: int | None = None,
) -> torch.Tensor:
    out = scatter_sum(src, index, dim, out, dim_size)
    dim_size = out.size(dim)

    index_dim = dim
    if index_dim < 0:
        index_dim = index_dim + src.dim()
    if index.dim() <= index_dim:
        index_dim = index.dim() - 1

    ones = torch.ones(index.size(), dtype=src.dtype, device=src.device)
    count = scatter_sum(ones, index, index_dim, None, dim_size)
    count[count < 1] = 1
    count = _broadcast(count, out, dim)
    if out.is_floating_point():
        out.true_divide_(count)
    else:
        out.div_(count, rounding_mode="floor")
    return out



import torch
import torch.nn.functional as F
from torch import Tensor

from chemeleon_dng.diffusion.models.base import DiffusionModelBase


class D3PM(DiffusionModelBase):
    """Discrete Denoising Diffusion Probabilistic Models (D3PM).
    """

    def __init__(
        self,
        *,
        betas: torch.Tensor,
        max_atoms: int,
        d3pm_hybrid_coeff: float = 1.0,
    ):
        self.betas = betas
        assert len(betas.shape) == 1, "betas must be 1-D"
        assert (betas > 0).all() and (betas <= 1).all()

        self.num_timesteps = int(betas.shape[0])
        self.max_atoms = max_atoms
        self.hybrid_coeff = d3pm_hybrid_coeff
        self.eps = 1.0e-6

        # transition matrix for absorbing
        self.q_one_step_mats = torch.stack(
            [
                self.get_absorbing_transition_mat(t)
                for t in range(0, self.num_timesteps)
            ],
            dim=0,
        )

        # construct transition matrices for q(x_t | x_start)
        q_mat_t = self.q_one_step_mats[0]
        q_mats = [q_mat_t]
        for t in range(0, self.num_timesteps):
            # Q_{1...t} = Q_{1 ... t-1} Q_t = Q_1 Q_2 ... Q_t
            q_mat_t = q_mat_t @ self.q_one_step_mats[t]
            q_mats.append(q_mat_t)
        q_mats = torch.stack(q_mats, dim=0)
        self.q_mats = q_mats

        assert self.q_mats.shape == (
            self.num_timesteps + 1,
            self.max_atoms,
            self.max_atoms,
        )

        self.q_one_step_transposed = self.q_one_step_mats.transpose(1, 2)

    def get_absorbing_transition_mat(self, t: int):
        """Computes transition matrix for q(x_t|x_{t-1}).

        Args:
            t (int): timestep.
            max_atoms (int): maximum number of atoms (100 + 1 for dummy atom).
                Defaults to 101.

        Returns:
            Q_t: transition matrix. shape = (max_atoms, max_atoms)
        """
        # get beta at timestep t
        beta_t = self.betas[t].float()

        diag = torch.full((self.max_atoms,), (1 - beta_t).item())
        mat = torch.diag(diag, 0)
        # add beta_t at first row
        mat[:, 0] += beta_t
        return mat

    def at(
        self,
        a: Tensor,
        t_per_node: Tensor,
        x: Tensor,
    ):
        """Extract coefficients at specified timesteps t - 1 and conditioning data x.

        Args:
            a (Tensor): matrix of coefficients. [num_timesteps, max_atoms, max_atoms]
            t_per_node (Tensor): timesteps.[B_n]
            x (Tensor): atom_types. [B_n]

        Returns:
            a[t, x] (Tensor): coefficients at timesteps t and data x. [B_n, max_atoms]
        """
        a = a.to(x.device)
        bs = t_per_node.shape[0]
        t_per_node = t_per_node.reshape((bs, *[1] * (x.dim() - 1)))
        # when t_per_node == 0, use t_per_node == 1
        t_per_node = torch.where(t_per_node == 0, 1, t_per_node)
        return a[t_per_node - 1, x, :]

    def q_sample(
        self,
        x_start: Tensor,
        t: Tensor,
        batch_idx: Tensor,
        noise: Tensor | None = None,
    ):
        """Sample from q(x_t | x_start) (i.e. add noise to the data).
        q(x_t | x_start) = Categorical(x_t ; p = x_start Q_{1...t}).

        Args:
            x_start (Tensor): Data at t=0.
            t (Tensor): Timesteps. [B_n]
            noise (Tensor): Noise. [B_n, max_atoms]
            batch_idx (Tensor): Batch index. [B_n]

        Returns:
            Tensor: [B_n, max_atoms]
        """
        t_per_node = t[batch_idx]
        logits = torch.log(self.at(self.q_mats, t_per_node, x_start) + self.eps)
        if noise is None:
            noise = torch.rand(len(batch_idx), self.max_atoms).to(
                x_start.device
            )  # uniform noise
        noise = torch.clip(noise, self.eps, 1.0)
        gumbel_noise = -torch.log(-torch.log(noise))
        return torch.argmax(logits + gumbel_noise, dim=-1)

    def q_posterior_logits(
        self,
        x_start: Tensor,
        x_t: Tensor,
        t_per_node: Tensor,
        is_x_start_one_hot: bool = False,
    ):
        """Compute logits for q(x_{t-1} | x_t, x_start)."""
        if is_x_start_one_hot:
            x_start_logits = x_start.clone()
        else:
            x_start_logits = torch.log(
                torch.nn.functional.one_hot(x_start, self.max_atoms) + self.eps
            )

        assert x_start_logits.shape == x_t.shape + (self.max_atoms,), print(
            f"x_start_logits.shape: {x_start_logits.shape}, x_t.shape: {x_t.shape}"
        )

        fact1 = self.at(self.q_one_step_transposed, t_per_node, x_t)

        softmaxed = torch.softmax(x_start_logits, dim=-1)
        self.q_mats = self.q_mats.to(x_start.device)
        qmats2 = self.q_mats[t_per_node - 2]
        fact2 = torch.einsum("b...c, bcd -> b...d", softmaxed, qmats2)

        out = torch.log(fact1 + self.eps) + torch.log(fact2 + self.eps)

        t_broadcast = t_per_node.reshape((t_per_node.shape[0], *[1] * (x_t.dim())))
        return torch.where(t_broadcast == 1, x_start_logits, out)

    def categorical_kl_logits(self, logits1, logits2, eps=1.0e-6):
        """KL divergence between categorical distributions.

        Distributions parameterized by logits.

        Args:
            logits1: logits of the first distribution. Last dim is class dim.
            logits2: logits of the second distribution. Last dim is class dim.
            eps: float small number to avoid numerical issues.

        Returns:
            KL(C(logits1) || C(logits2)): shape: logits1.shape[:-1]
        """
        out = torch.softmax(logits1 + eps, dim=-1) * (
            torch.log_softmax(logits1 + eps, dim=-1)
            - torch.log_softmax(logits2 + eps, dim=-1)
        )
        return out.sum(dim=-1).mean()

    @torch.no_grad()
    def p_sample(
        self,
        pred_x_start_logits: Tensor,
        x_t: Tensor,
        t: Tensor,
        batch_idx: Tensor,
        noise: Tensor | None = None,
    ):
        t_per_node = t[batch_idx]
        pred_q_posterior_logits = self.q_posterior_logits(
            pred_x_start_logits, x_t, t_per_node, is_x_start_one_hot=True
        )

        if noise is None:
            noise = torch.rand(len(batch_idx), self.max_atoms).to(
                x_t.device
            )  # uniform noise
        noise = torch.clamp(noise, min=self.eps, max=1.0)
        # if t == 1, use x_start_logits
        nonzero_mask = (t_per_node != 1).to(x_t.dtype).view(-1, *([1] * (x_t.ndim)))
        gumbel_noise = -torch.log(-torch.log(noise))
        pred_sample = torch.argmax(
            pred_q_posterior_logits + gumbel_noise * nonzero_mask, dim=-1
        )
        return pred_sample

    def training_loss(
        self,
        *,
        pred_x_start_logits: Tensor,
        x_start: Tensor,
        x_t: Tensor,
        t: Tensor,
        batch_idx: Tensor,
    ):
        t_per_node = t[batch_idx]
        true_q_posterior_logits = self.q_posterior_logits(x_start, x_t, t_per_node)
        pred_q_posterior_logits = self.q_posterior_logits(
            pred_x_start_logits, x_t, t_per_node, is_x_start_one_hot=True
        )
        vb_loss = self.categorical_kl_logits(
            true_q_posterior_logits, pred_q_posterior_logits
        )
        ce_loss = F.cross_entropy(pred_x_start_logits.flatten(0, -2), x_start.flatten())
        loss = vb_loss + ce_loss * self.hybrid_coeff
        return {
            "loss": loss,
            "vb_loss": vb_loss,
            "ce_loss": ce_loss,
        }



def create_model(
    *,
    task,
    hidden_dim,
    time_dim,
    num_layers,
    max_atoms,
    act_fn,
    dis_emb,
    num_freqs,
    ln,
    ip,
    smooth,
    cond_dim,
    pred_atom_types,
):
    from chemeleon_dng.cspnet import CSPNet, CSPNetTask

    if task.lower() == "csp":
        task = CSPNetTask.CSP
        assert pred_atom_types is False, cond_dim <= 0
        cond_dim = 0
    elif task.lower() == "dng":
        task = CSPNetTask.DNG
        assert pred_atom_types is True, cond_dim <= 0
    elif task.lower() == "guide":
        task = CSPNetTask.GUIDE
        assert pred_atom_types is True, cond_dim > 0
    elif task.lower() == "clip":
        task = CSPNetTask.CLIP
        assert pred_atom_types is False, cond_dim <= 0
    else:
        raise ValueError(f"Unknown task {task}")
    model = CSPNet(
        hidden_dim=hidden_dim,
        time_dim=time_dim,
        num_layers=num_layers,
        max_atoms=max_atoms,
        act_fn=act_fn,
        dis_emb=dis_emb,
        num_freqs=num_freqs,
        ln=ln,
        ip=ip,
        smooth=smooth,
        cond_dim=cond_dim,
        pred_atom_types=pred_atom_types,
        task=task,
    )
    return model


def create_diffusion_module(
    *,
    task,
    model_configs,
    optimizer_configs,
    num_timesteps,
    # ddpm
    beta_schedule_ddpm,
    # d3pm
    beta_schedule_d3pm,
    max_atoms,
    d3pm_hybrid_coeff,
    # dsm
    sigma_begin,
    sigma_end,
):
    from chemeleon_dng.diffusion.diffusion_module import (
        DiffusionModule,
        DiffusionModuleTask,
    )
    from chemeleon_dng.diffusion.diffusion_scheduler import get_named_beta_schedule
    from chemeleon_dng.diffusion.models.d3pm import D3PM
    from chemeleon_dng.diffusion.models.ddpm import DDPM
    from chemeleon_dng.diffusion.models.dsm import DSM

    if task.lower() == "csp":
        task = DiffusionModuleTask.CSP
        model_configs["pred_atom_types"] = False
    elif task.lower() == "dng":
        task = DiffusionModuleTask.DNG
        model_configs["pred_atom_types"] = True
    elif task.lower() == "guide":
        task = DiffusionModuleTask.GUIDE
        model_configs["pred_atom_types"] = True
    else:
        raise ValueError("DiffusionModuleTask should be in `csp`, `dng`, `guide`.")
    betas_d3pm = get_named_beta_schedule(beta_schedule_d3pm, num_timesteps)
    d3pm = D3PM(
        betas=betas_d3pm, max_atoms=max_atoms, d3pm_hybrid_coeff=d3pm_hybrid_coeff
    )
    betas_ddpm = get_named_beta_schedule(beta_schedule_ddpm, num_timesteps)
    ddpm = DDPM(betas=betas_ddpm)
    dsm = DSM(num_timesteps=num_timesteps, sigma_begin=sigma_begin, sigma_end=sigma_end)
    # Set up DiffusionModule
    diffusion_module = DiffusionModule(
        task=task,
        num_timesteps=num_timesteps,
        diffusion_atom_type=d3pm if task != DiffusionModuleTask.CSP else None,
        diffusion_lattice=ddpm,
        diffusion_frac_coord=dsm,
        model_configs=model_configs,
        optimizer_configs=optimizer_configs,
    )
    return diffusion_module


def create_reinforce_module(
    *,
    diffusion_module_path,
    optimizer_configs,
    reward_type,
    rl_algorithm,
    num_group_samples,
    num_inner_batches,
    clip_ratio,
    kl_coeff,
    weight_atom_type,
    weight_lattice,
    weight_frac_coord,
    group_reward_normalize,
    reward_normalize,
):
    from chemeleon_dng.reinforce.reinforce_module import ReinforceModule

    reinforce_module = ReinforceModule(
        diffusion_module_path=diffusion_module_path,
        reward_type=reward_type,
        rl_algorithm=rl_algorithm,
        num_group_samples=num_group_samples,
        num_inner_batches=num_inner_batches,
        clip_ratio=clip_ratio,
        kl_coeff=kl_coeff,
        weight_atom_type=weight_atom_type,
        weight_lattice=weight_lattice,
        weight_frac_coord=weight_frac_coord,
        group_reward_normalize=group_reward_normalize,
        reward_normalize=reward_normalize,
        optimizer_configs=optimizer_configs,
    )
    return reinforce_module



import torch
import torch.nn.functional as F
from torch import Tensor

from chemeleon_dng.diffusion.models.base import DiffusionModelBase


class DDPM(DiffusionModelBase):
    """Denoising Diffusion Probabilistic Model (DDPM).
    """

    def __init__(
        self,
        *,
        betas: torch.Tensor,
    ):
        betas = torch.cat((torch.tensor([0.0]), betas))
        assert len(betas.shape) == 1, "betas must be 1-D"

        self.num_timesteps = int(betas.shape[0])

        self.alphas = 1.0 - betas
        self.alphas_cumprod = torch.cumprod(self.alphas, dim=0)

        # Coefficients for diffusion q(x_t | x_{t-1}) and others
        self.sqrt_alphas_cumprod = torch.sqrt(self.alphas_cumprod)
        self.sqrt_one_minus_alphas_cumprod = torch.sqrt(1.0 - self.alphas_cumprod)

        # Coefficients for posterior q(x_{t-1} | x_t, x_0)
        self.posterior_variance = torch.zeros_like(betas)
        self.posterior_variance[1:] = (
            betas[1:]
            * (1.0 - self.alphas_cumprod[:-1])
            / (1.0 - self.alphas_cumprod[1:])
        )
        # DDPM Coefficients
        self.ddpm_posterior_mean_coef1 = torch.ones_like(betas)
        self.ddpm_posterior_mean_coef1[1:] = 1.0 / torch.sqrt(self.alphas[1:])
        self.ddpm_posterior_mean_coef2 = torch.zeros_like(betas)
        self.ddpm_posterior_mean_coef2[1:] = (1.0 - self.alphas[1:]) / torch.sqrt(
            1 - self.alphas_cumprod[1:]
        )

    def q_sample(self, x_start, t, noise=None):
        """Sample from the forward process q(x_t | x_0).

        :param x_start: the initial data batch.
        :param t: the number of diffusion steps.
        :param noise: if specified, the split-out normal noise.
        """
        if noise is None:
            noise = torch.randn_like(x_start)

        assert noise.shape == x_start.shape
        return (
            _extract_into_tensor(self.sqrt_alphas_cumprod, t, x_start.shape) * x_start
            + _extract_into_tensor(self.sqrt_one_minus_alphas_cumprod, t, x_start.shape)
            * noise
        )

    @torch.no_grad()
    def p_sample(self, model_output: Tensor, x_t: Tensor, t: Tensor):
        """Sample x_{t-1} from the model at the given timestep.

        :param model_output: the model output, which is either the predicted epsilon.
        :param x_t: the current tensor at x_{t}.
        :param t: the value of t, starting at 0 for the first diffusion step.
        """
        noise = torch.randn_like(x_t)
        nonzero_mask = (
            (t != 1).float().view(-1, *([1] * (len(x_t.shape) - 1)))
        )  # no noise when t == 1

        c0 = _extract_into_tensor(self.ddpm_posterior_mean_coef1, t, x_t.shape)
        c1 = _extract_into_tensor(self.ddpm_posterior_mean_coef2, t, x_t.shape)
        posterior_std = _extract_into_tensor(
            torch.sqrt(self.posterior_variance), t, x_t.shape
        )
        pred_sample = (
            c0 * (x_t - c1 * model_output) + nonzero_mask * noise * posterior_std
        )
        return pred_sample

    def training_loss(
        self,
        *,
        model_output: Tensor,
        noise: Tensor,
    ):
        loss = F.mse_loss(model_output, noise)  # Model mean type = epsilon
        return {"loss": loss}

    def p_mean_and_var(self, model_output: Tensor, x_t: Tensor, t: Tensor):
        """Compute the mean and variance of the posterior distribution.

        :param model_output: the model output, which is either the predicted epsilon.
        :param x_t: the current tensor at x_{t}.
        :param t: the value of t, starting at 0 for the first diffusion step.
        """
        c0 = _extract_into_tensor(self.ddpm_posterior_mean_coef1, t, x_t.shape)
        c1 = _extract_into_tensor(self.ddpm_posterior_mean_coef2, t, x_t.shape)
        var = _extract_into_tensor(self.posterior_variance, t, x_t.shape)
        mean = c0 * (x_t - c1 * model_output)
        return mean, var


def _extract_into_tensor(arr, timesteps, broadcast_shape):
    """Extract values from a 1-D numpy array for a batch of indices.

    :param arr: the 1-D numpy array.
    :param timesteps: a tensor of indices into the array to extract.
    :param broadcast_shape: a larger shape of K dimensions with the batch
                            dimension equal to the length of timesteps.
    :return: a tensor of shape [batch_size, 1, ...] where the shape has K dims.
    """
    arr = arr.to(timesteps.device)
    res = arr[timesteps].float()
    while len(res.shape) < len(broadcast_shape):
        res = res[..., None]
    return res.expand(broadcast_shape)



# https://github.com/jiaor17/DiffCSP
import enum
import math
from collections import namedtuple

import torch
import torch.nn as nn
from torch_geometric.utils import dense_to_sparse

from chemeleon_dng.scatter import scatter_mean


class CSPNetTask(enum.Enum):
    """Which task of the model.
    """

    CSP = enum.auto()  # Crystal Structure Prediction
    DNG = enum.auto()  # De Novo Generation
    GUIDE = enum.auto()  # Guiding Generation
    CLIP = enum.auto()  # Contrastive Learning


CSPNET_OUTPUTS = namedtuple(
    "DECODER_OUTPUTS", ["atom_types", "lattices", "frac_coords", "node_features"]
)


class SinusoidalTimeEmbeddings(nn.Module):
    """Attention is all you need."""

    def __init__(self, dim):
        super().__init__()
        self.dim = dim

    def forward(self, time):
        device = time.device
        half_dim = self.dim // 2
        embeddings = math.log(10000) / (half_dim - 1)
        embeddings = torch.exp(torch.arange(half_dim, device=device) * -embeddings)
        embeddings = time[:, None] * embeddings[None, :]
        embeddings = torch.cat((embeddings.sin(), embeddings.cos()), dim=-1)
        return embeddings


class SinusoidsEmbedding(nn.Module):
    """Embedding for periodic distance features."""

    def __init__(self, n_frequencies=10, n_space=3):
        super().__init__()
        self.n_frequencies = n_frequencies
        self.n_space = n_space
        self.frequencies = 2 * math.pi * torch.arange(self.n_frequencies)
        self.dim = self.n_frequencies * 2 * self.n_space

    def forward(self, x):
        emb = x.unsqueeze(-1) * self.frequencies[None, None, :].to(x.device)
        emb = emb.reshape(-1, self.n_frequencies * self.n_space)
        emb = torch.cat((emb.sin(), emb.cos()), dim=-1)
        return emb.detach()


class FilmLayer(nn.Module):
    """FiLM layer for efficient incorporation of conditional embeddings."""

    def __init__(
        self,
        hidden_dim,
        cond_dim,
        act_fn=nn.SiLU(),
    ):
        super(FilmLayer, self).__init__()
        self.hidden_dim = hidden_dim
        self.cond_dim = cond_dim
        self.act_fn = act_fn
        self.mlp_cond = nn.Sequential(
            nn.Linear(cond_dim, hidden_dim * 2),
            act_fn,
        )
        # block
        self.proj = nn.Linear(hidden_dim, hidden_dim)
        self.norm = nn.LayerNorm(hidden_dim)

    def forward(self, *, x, cond_embeds):
        assert cond_embeds is not None
        cond_emb = self.mlp_cond(cond_embeds)
        scale, shift = cond_emb.chunk(2, dim=1)
        # residual block
        x_init = x.clone()
        x = self.proj(x)
        x = self.norm(x)
        x = x * scale + shift
        x = self.act_fn(x)
        x = x + x_init
        return x


class CSPLayer(nn.Module):
    """Message passing layer for cspnet."""

    def __init__(
        self, hidden_dim=128, act_fn=nn.SiLU(), dis_emb=None, ln=False, ip=True
    ):
        super(CSPLayer, self).__init__()

        self.dis_dim = 3
        self.dis_emb = dis_emb
        self.ip = ip
        if dis_emb is not None:
            self.dis_dim = dis_emb.dim
        self.edge_mlp = nn.Sequential(
            nn.Linear(hidden_dim * 2 + 9 + self.dis_dim, hidden_dim),
            act_fn,
            nn.Linear(hidden_dim, hidden_dim),
            act_fn,
        )
        self.node_mlp = nn.Sequential(
            nn.Linear(hidden_dim * 2, hidden_dim),
            act_fn,
            nn.Linear(hidden_dim, hidden_dim),
            act_fn,
        )
        self.ln = ln
        if self.ln:
            self.layer_norm = nn.LayerNorm(hidden_dim)

    def edge_model(
        self,
        node_features,
        frac_coords,
        lattices,
        edge_index,
        edge2graph,
        frac_diff=None,
    ):
        hi, hj = node_features[edge_index[0]], node_features[edge_index[1]]
        if frac_diff is None:
            xi, xj = frac_coords[edge_index[0]], frac_coords[edge_index[1]]
            frac_diff = (xj - xi) % 1.0
        if self.dis_emb is not None:
            frac_diff = self.dis_emb(frac_diff)  # fourier transform
        if self.ip:
            lattice_ips = lattices @ lattices.transpose(-1, -2)
        else:
            lattice_ips = lattices
        lattice_ips_flatten = lattice_ips.view(-1, 9)
        lattice_ips_flatten_edges = lattice_ips_flatten[edge2graph]
        edges_input = torch.cat([hi, hj, lattice_ips_flatten_edges, frac_diff], dim=1)
        edge_features = self.edge_mlp(edges_input)
        return edge_features

    def node_model(self, node_features, edge_features, edge_index):
        agg = scatter_mean(
            edge_features,
            edge_index[0],
            dim=0,
            dim_size=node_features.shape[0],
        )
        agg = torch.cat([node_features, agg], dim=1)
        out = self.node_mlp(agg)
        return out

    def forward(
        self,
        *,
        node_features,
        lattices,
        frac_coords,
        edge_index,
        edge2graph,
        frac_diff=None,
    ):
        node_input = node_features
        if self.ln:
            node_features = self.layer_norm(node_input)
        edge_features = self.edge_model(
            node_features, frac_coords, lattices, edge_index, edge2graph, frac_diff
        )
        node_output = self.node_model(node_features, edge_features, edge_index)
        return node_input + node_output


class CSPNet(nn.Module):
    """CSPNet model, adopted from DiffCSP.

    - edge_style = fc
    - Task
    (1) CSP: pred_atom_types = False
    (2) DNG: pred_atom_types = True
    (3) Guide: cond_dim != 0
    (4) CLIP: time_dim = 0, cond_dim = 0
    """

    def __init__(
        self,
        hidden_dim=512,
        time_dim=128,
        num_layers=6,
        max_atoms=100,
        act_fn="silu",
        dis_emb="sin",
        num_freqs=128,
        ln=True,
        ip=True,
        smooth=False,
        cond_dim=0,
        pred_atom_types=True,
        task: CSPNetTask | None = None,
    ):
        super().__init__()
        self.task = task
        self.hidden_dim = hidden_dim
        self.time_dim = time_dim
        self.cond_dim = cond_dim
        self.pred_atom_types = pred_atom_types

        # Validate the configuration matches the task
        if task is not None:
            self._validate_task(task, time_dim, cond_dim, pred_atom_types)

        # Time embeddings
        if time_dim > 0:
            self.time_embedding = SinusoidalTimeEmbeddings(time_dim)
        else:
            self.time_embedding = None

        # Node embedding
        if smooth:
            self.node_embedding = nn.Linear(max_atoms, hidden_dim)
        else:
            self.node_embedding = nn.Embedding(max_atoms, hidden_dim)

        # Atom latent embedding (node + time embedding)
        if time_dim > 0:
            self.atom_latent_embedding = nn.Linear(hidden_dim + time_dim, hidden_dim)

        # Set up FiLM layer if time or condition embeddings are used
        if cond_dim > 0:
            self.film_layer = FilmLayer(hidden_dim, cond_dim)
        else:
            self.film_layer = None

        # Set up activation function
        if act_fn == "silu":
            self.act_fn = nn.SiLU()

        # Set up distance embedding
        if dis_emb == "sin":
            self.dis_emb = SinusoidsEmbedding(n_frequencies=num_freqs)
        elif dis_emb == "none":
            self.dis_emb = None

        # Set up layers
        self.num_layers = num_layers
        self.ln = ln
        self.ip = ip
        for i in range(0, num_layers):
            self.add_module(
                f"csp_layer_{i}",
                CSPLayer(hidden_dim, self.act_fn, self.dis_emb, ln=ln, ip=ip),
            )
        if ln:
            self.final_layer_norm = nn.LayerNorm(hidden_dim)
        if self.pred_atom_types:
            self.type_out = nn.Linear(hidden_dim, max_atoms)
        self.coord_out = nn.Linear(hidden_dim, 3, bias=False)
        self.lattice_out = nn.Linear(hidden_dim, 9, bias=False)

    def _validate_task(
        self, task: CSPNetTask, time_dim: int, cond_dim: int, pred_atom_types: bool
    ):
        if task == CSPNetTask.CSP:
            if pred_atom_types:
                raise ValueError("CSP task requires pred_atom_types = False")
        elif task == CSPNetTask.DNG:
            if not pred_atom_types:
                raise ValueError("DNG task requires pred_atom_types = True")
        elif task == CSPNetTask.GUIDE:
            if cond_dim <= 0 or not pred_atom_types:
                raise ValueError(
                    "GUIDE task requires cond_dim > 0 and pred_atom_types = True"
                )
        elif task == CSPNetTask.CLIP:
            if time_dim != 0 or cond_dim != 0 or pred_atom_types:
                raise ValueError(
                    "CLIP task requires time_dim = 0, cond_dim = 0, and pred_atom_types = False"
                )
        else:
            raise ValueError(f"Unsupported task: {task}")

    def gen_edges(self, num_atoms, frac_coords):
        # edge_style = fc
        lis = [torch.ones(n, n, device=num_atoms.device) for n in num_atoms]
        fc_graph = torch.block_diag(*lis)
        fc_edges, _ = dense_to_sparse(fc_graph)
        return fc_edges, (frac_coords[fc_edges[1]] - frac_coords[fc_edges[0]]) % 1.0

    def forward(
        self,
        *,
        atom_types,
        lattices,
        frac_coords,
        num_atoms,
        batch_idx,
        t=None,  # time embeddings can be omitted when training contrastive learning
        cond_embeds=None,
    ) -> CSPNET_OUTPUTS:
        edges, frac_diff = self.gen_edges(num_atoms, frac_coords)
        edge2graph = batch_idx[edges[0]]
        node_features = self.node_embedding(atom_types)  # [B_n, hidden_dim]

        if t is not None:
            assert self.time_embedding is not None, "Time embeddings are not defined."
            time_embeds = self.time_embedding(t)  # [B, time_dim]
            time_embeds_per_atom = time_embeds.repeat_interleave(
                num_atoms, dim=0
            )  # [B_n, time_dim]
            # Concatenate time embeddings with node features
            node_features = torch.cat(
                [node_features, time_embeds_per_atom], dim=1
            )  # [B_n, hidden_dim + time_dim]
            node_features = self.atom_latent_embedding(
                node_features
            )  # [B_n, hidden_dim]

        if cond_embeds is not None:
            assert self.film_layer is not None, "FiLM layer is not defined."
            cond_embeds_per_atom = cond_embeds.repeat_interleave(
                num_atoms, dim=0
            )  # [B_n, cond_dim]
        else:
            cond_embeds_per_atom = None

        for i in range(0, self.num_layers):
            if self.film_layer is not None:
                node_features = self.film_layer(
                    x=node_features, cond_embeds=cond_embeds_per_atom
                )
            node_features = self._modules[f"csp_layer_{i}"](
                node_features=node_features,
                lattices=lattices,
                frac_coords=frac_coords,
                edge_index=edges,
                edge2graph=edge2graph,
                frac_diff=frac_diff,
            )  # type: ignore

        if self.ln:
            node_features = self.final_layer_norm(node_features)

        coord_out = self.coord_out(node_features)

        graph_features = scatter_mean(node_features, batch_idx, dim=0)
        lattice_out = self.lattice_out(graph_features)
        lattice_out = lattice_out.view(-1, 3, 3)
        if self.ip:
            lattice_out = torch.einsum("bij,bjk->bik", lattice_out, lattices)
        if self.pred_atom_types:
            type_out = self.type_out(node_features)
        else:
            type_out = atom_types

        return CSPNET_OUTPUTS(
            atom_types=type_out,  # [B_n, max_atoms]
            lattices=lattice_out,  # [B, 3, 3]
            frac_coords=coord_out,  # [B_n, 3]
            node_features=node_features,  # [B_n, hidden_dim]
        )



"""This script samples crystal structures using a trained diffusion model.
It supports three types of tasks:
- CSP (Crystal Structure Prediction): Predicts stable crystal structures from given atom types
- DNG (De Novo Generation): Generates new crystal structures from scratch
- GUIDE (Guiding Generation): Generates structures with guidance.
"""

import json
from pathlib import Path

import fire
import numpy as np
import torch
from monty.serialization import dumpfn
from pymatgen.core import Composition, Structure

from chemeleon_dng.dataset.num_atom_distributions import NUM_ATOM_DISTRIBUTIONS
from chemeleon_dng.diffusion.diffusion_module import DiffusionModule
from chemeleon_dng.download_util import get_checkpoint_path

DEFAULT_MODEL_PATH = {
    "csp": "ckpts/chemeleon_csp_alex_mp_20_v0.0.2.ckpt",
    "dng": "ckpts/chemeleon_dng_alex_mp_20_v0.0.2.ckpt",
    "guide": ".",
}


def sample_csp(
    dm: DiffusionModule,
    formulas: str | tuple,
    num_samples: int,
    batch_size: int,
    output_path: Path,
):
    # Parse formulas
    if isinstance(formulas, (tuple, list)):
        formula_list = [f.strip() for f in formulas]
    else:
        formula_list = [formulas.strip()]
    print(f"Generating {num_samples} samples for each formula: {formula_list}")

    # Generate samples in batches
    def repeat_elements(elements, times):
        return [x for x in elements for _ in range(times)]

    total_formula_list = repeat_elements(formula_list, num_samples)
    for i in range(0, len(total_formula_list), batch_size):
        print(f"Generating batch #{i // batch_size + 1} with {batch_size} samples.")
        batch_formula_list = total_formula_list[i : i + batch_size]

        atom_types = []
        num_atoms = []
        for formula in batch_formula_list:
            comp = Composition(formula)
            atomic_numbers = [el.Z for el, amt in comp.items() for _ in range(int(amt))]
            atom_types.extend(atomic_numbers)
            num_atoms.append(len(atomic_numbers))

        assert len(atom_types) == sum(num_atoms)
        gen_atoms_list = dm.sample(
            task="csp", atom_types=atom_types, num_atoms=num_atoms
        )

        # Save generated structures
        for j, atoms in enumerate(gen_atoms_list):
            atoms.write(output_path / f"sample_{i+j}_{atoms.get_chemical_formula()}.cif")  # type: ignore


def sample_dng(
    dm: DiffusionModule,
    num_atom_distribution: str | list[int],
    num_samples: int,
    batch_size: int,
    output_path: Path,
):
    if isinstance(num_atom_distribution, str):
        atom_distribution = NUM_ATOM_DISTRIBUTIONS[num_atom_distribution]
        num_atoms = np.random.choice(
            list(atom_distribution.keys()),
            p=list(atom_distribution.values()),
            size=num_samples,
        ).tolist()
    elif isinstance(num_atom_distribution, list):
        num_atoms = num_atom_distribution
        assert (
            len(num_atoms) == num_samples
        ), "num_atom_distribution must be a list of length num_samples."
    else:
        raise ValueError(
            "num_atom_distribution must be either a string or a list of integers."
        )
    print(
        f"Generating {num_samples} samples with atom distributions: {num_atom_distribution}"
    )

    # Generate samples in batches
    for i in range(0, num_samples, batch_size):
        print(f"Generating batch #{i // batch_size + 1} with {batch_size} samples.")
        batch_num_atoms = num_atoms[i : i + batch_size]
        gen_atoms_list = dm.sample(task="dng", num_atoms=batch_num_atoms)

        # Save generated structures
        for j, atoms in enumerate(gen_atoms_list):
            atoms.write(output_path / f"sample_{i+j}_{atoms.get_chemical_formula()}.cif")  # type: ignore


def sample(
    task: str,
    num_samples: int = 100,
    batch_size: int | None = None,
    formulas: str | tuple | list | None = None,  # Only for CSP task
    num_atom_distribution: str | list[int] | None = "mp-20",  # Only for DNG task
    model_path: str | None = None,
    output_dir: str = "./results",
    device: str | None = None,
    save_json: bool = True,
):
    """Sample crystal structures using a trained diffusion model.

    :param task: Type of sampling task ('csp', 'dng', or 'guide').
                 'csp' for Crystal Structure Prediction,
                 'dng' for De Novo Generation,
                 'guide' for Guided Generation.
    :param num_samples: Number of samples to generate per formula (for CSP) or total samples (for DNG), defaults to 100
    :param batch_size: Number of samples to generate per batch, defaults to num_samples if None
    :param formulas: Chemical formula(s) for crystal structure prediction (CSP task).
                    Can be a single string like 'NaCl' or a tuple of strings like ('NaCl', 'SiO2').
                    Required for CSP task, defaults to None
    :param num_atom_distribution: Distribution of number of atoms per unit cell for DNG task.
                                 Can be a predefined distribution name (e.g., 'mp-20') or
                                 a list of integers specifying the number of atoms for each sample.
                                 Required for DNG task, defaults to "mp-20"
    :param model_path: Path to the checkpoint file of the trained model.
                     If None, uses the default checkpoint for the specified task, defaults to None
    :param output_dir: Directory to save the generated structures, defaults to "./results"
    :param device: Device to run the model on ('cuda' or 'cpu').
                  If None, uses CUDA if available, otherwise CPU, defaults to None
    :param save_json: Whether to save the generated structures in JSON format, defaults to True

    Examples:
    --------
    # CSP: Generate 100 samples each for NaCl, OCl3, and UO15 formulas, processing 150 samples at a time
    >>> sample(task="csp", formulas="NaCl, OCl3, UO15", num_samples=100, batch_size=150, output_dir="./results")
    or
    >>> sample(task="csp", formulas=["NaCl", "OCl3", "UO15"], num_samples=100, batch_size=150, output_dir="./results")

    # CSP: Generate 100 samples for a single formula at a time
    >>> sample(task="csp", formulas="SiO2", num_samples=100)

    # DNG: Generate 100 samples with specific atom counts [7,8]
    >>> sample(task="dng", num_atom_distribution=[7, 8] * 50, num_samples=100)

    # DNG: Generate 100 samples using the 'mp-20' atom distribution
    >>> sample(task="dng", num_samples=100)

    # Using a custom checkpoint path
    >>> sample(task="csp", formulas="NaCl", model_path="path/to/custom/checkpoint.ckpt")

    # Explicitly specify device
    >>> sample(task="dng", device="cpu", num_samples=5)
    """
    # Set device
    if device is None:
        device = "cuda" if torch.cuda.is_available() else "cpu"
    print(f"Using device: {device}")

    # Set default batch size if not provided
    if batch_size is None:
        batch_size = num_samples

    # Set default checkpoint path if not provided
    if model_path is None:
        model_path = get_checkpoint_path(task, DEFAULT_MODEL_PATH)
    print(f"Using checkpoint path: {model_path}")

    # Load the diffusion module
    dm = DiffusionModule.load_from_checkpoint(model_path, map_location=device)

    # Validate the checkpoint’s task matches requested task
    assert (
        dm.model.task.name.lower() == task.lower()  # type: ignore
    ), f"Checkpoint task does not match the provided task '{task}'."

    # Set Output Directory
    output_path = Path(output_dir)
    output_path.mkdir(parents=True, exist_ok=True)
    print(f"The generated structures will be saved in: {output_path}")
    existing_files = list(output_path.glob("sample_*.cif"))
    if existing_files:
        print(
            f"Warning: {len(existing_files)} existing CIF files found in the output directory."
            " Recommend emptying the directory before sampling to avoid overwriting."
        )

    # Validate task
    task = task.lower()
    if task == "csp":
        assert formulas is not None, "Formulas must be provided for CSP task."
        if isinstance(formulas, str):
            if formulas.endswith(".json"):
                with open(formulas) as f:
                    formulas = json.load(f)
            else:
                formulas = tuple(formulas.split(","))
        sample_csp(
            dm=dm,
            formulas=formulas,
            num_samples=num_samples,
            batch_size=batch_size,
            output_path=output_path,
        )
    elif task == "dng":
        assert (
            num_atom_distribution is not None
        ), "num_atom_distribution must be provided for DNG task."
        sample_dng(
            dm=dm,
            num_atom_distribution=num_atom_distribution,
            num_samples=num_samples,
            batch_size=batch_size,
            output_path=output_path,
        )
    elif task == "guide":
        assert (
            dm.model.task.name.lower() == task  # type: ignore
        ), "Checkpoint task does not match the provided task."
    else:
        raise ValueError(
            f"Unsupported task: {task}. Supported tasks are: csp, dng, guide."
        )

    # Save generated structures in JSON format
    if save_json:
        gen_atoms_files = list(output_path.glob("sample_*.cif"))
        all_gen_atoms_list = [Structure.from_file(file) for file in gen_atoms_files]
        dumpfn(all_gen_atoms_list, output_path / "generated_structures.json.gz")
        print(
            f"The {len(all_gen_atoms_list)} generated structures saved in JSON format at: {output_path / 'generated_structures.json.gz'}"
        )


if __name__ == "__main__":
    fire.Fire(sample)



from abc import abstractmethod

import torch
from pytorch_lightning import LightningModule
from torch import Tensor
from torch.optim.adam import Adam
from torch.optim.adamw import AdamW
from torch.optim.sgd import SGD

from chemeleon_dng.schema import CrystalBatch


class BaseModule(LightningModule):
    """Base module for configuration of optimizers and logging.
    """

    def __init__(
        self,
        optimizer: str,
        lr: float,
        weight_decay: float,
        scheduler: str,
        patience: int,
        early_stopping: int,
        warmup_steps: int = 0,
    ):
        super().__init__()
        self.optimizer = optimizer
        self.lr = lr
        self.weight_decay = weight_decay
        self.scheduler = scheduler
        self.patience = patience
        self.early_stopping = early_stopping
        self.warmup_steps = warmup_steps

    @abstractmethod
    def calculate_loss(self, *args, **kwargs) -> dict:
        """Calculate loss for the BaseModule.

        The output should be a dictionary, the key "total_loss" must be included.
        The value of key "total_loss" will be used for backpropagation.
        The other keys will be used for logging.
        """
        raise NotImplementedError("calculate_loss is not implemented")

    def training_step(
        self, batch: CrystalBatch, batch_idx: int  # pylint: disable=unused-argument
    ) -> Tensor:
        res = self.calculate_loss(batch)
        self._log_metrics(res=res, split="train", batch_size=batch.num_graphs)
        return res["total_loss"]

    def validation_step(
        self, batch: CrystalBatch, batch_idx: int  # pylint: disable=unused-argument
    ) -> Tensor:
        res = self.calculate_loss(batch)
        self._log_metrics(res=res, split="val", batch_size=batch.num_graphs)
        return res["total_loss"]

    def configure_optimizers(self):
        return self._set_configure_optimizers()

    @torch.no_grad()
    def _log_metrics(
        self,
        res: dict,
        split: str,
        batch_size: int | None = None,
    ):
        for k, v in res.items():
            self.log(
                f"{split}/{k}",
                v,
                batch_size=batch_size,
                on_step=True if split == "train" else False,
                sync_dist=True,
                prog_bar=True,
            )

    def _set_configure_optimizers(self):
        # Set optimizer
        if self.optimizer == "adam":
            print(f"Using Adam optimizer with lr={self.lr}")
            optimizer = Adam(
                self.parameters(), lr=self.lr, weight_decay=self.weight_decay
            )
        elif self.optimizer == "sgd":
            optimizer = SGD(
                self.parameters(), lr=self.lr, weight_decay=self.weight_decay
            )
        elif self.optimizer == "adamw":
            optimizer = AdamW(
                self.parameters(), lr=self.lr, weight_decay=self.weight_decay
            )
        else:
            raise ValueError(f"Invalid optimizer: {self.optimizer}")
        # Get max_steps
        if self.trainer.max_steps == -1:
            max_steps = self.trainer.estimated_stepping_batches
        else:
            max_steps = self.trainer.max_steps
        # Set scheduler
        if self.scheduler == "constant":
            scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, lambda _: 1.0)
        elif self.scheduler == "cosine":
            scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max=10)
        elif self.scheduler == "reduce_on_plateau":
            scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
                optimizer, mode="min", min_lr=1e-4, factor=0.6, patience=self.patience
            )
        elif self.scheduler == "linear_decay":
            scheduler = torch.optim.lr_scheduler.LinearLR(optimizer, max_steps)
        elif self.scheduler == "warmup_linear":
            assert self.warmup_steps > 0, "warmup_steps must be greater than 0"
            if self.warmup_steps < 1.0:
                self.warmup_steps = int(max_steps * self.warmup_steps)
            scheduler = torch.optim.lr_scheduler.LinearLR(
                optimizer,
                start_factor=0.01,
                end_factor=1,
                total_iters=self.warmup_steps,
            )
        else:
            raise ValueError(f"Invalid scheduler: {self.scheduler}")
        lr_scheduler = {
            "scheduler": scheduler,
            "name": "learning rate",
            "monitor": "val/total_loss",
            "frequency": self.trainer.check_val_every_n_epoch,
        }

        return ([optimizer], [lr_scheduler])



from pytorch_lightning import LightningDataModule
from torch_geometric.loader import DataLoader

from chemeleon_dng.dataset.mp_dataset import MPDataset


class DataModule(LightningDataModule):
    """DataModule for ChemeleonRL."""

    def __init__(
        self,
        data_dir: str,
        batch_size: int,
        dataset_type: str = "mp",
        target_condition: str | None = None,
        num_workers: int = 0,
        pin_memory: bool = True,
        val_batch_size: int | None = None,
        test_batch_size: int | None = None,
    ):
        super().__init__()
        # Configs for dataset
        self.dataset_type = dataset_type
        self.data_dir = data_dir
        self.target_condition = target_condition
        print(f"Data directory: {self.data_dir}")

        # Configs for dataloader
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.pin_memory = pin_memory
        self.val_batch_size = batch_size if val_batch_size is None else val_batch_size
        self.test_batch_size = (
            batch_size if test_batch_size is None else test_batch_size
        )

        # Initialize datasets
        self.train_dataset = None
        self.val_dataset = None
        self.test_dataset = None

    @property
    def dataset_cls(self):
        """Return the dataset class based on the dataset type."""
        if self.dataset_type == "mp":
            return MPDataset
        elif self.dataset_type == "mp_20":
            return MPDataset
        elif self.dataset_type == "mp_alex_20":
            return MPDataset
        else:
            raise ValueError(f"Unknown dataset type: {self.dataset_type}")

    def setup(self, stage: str | None = None):
        if stage == "fit" or stage is None:
            self.train_dataset = self.dataset_cls(
                data_dir=self.data_dir,
                split="train",
                target_condition=self.target_condition,
            )
            self.val_dataset = self.dataset_cls(
                data_dir=self.data_dir,
                split="val",
                target_condition=self.target_condition,
            )
        if stage == "test" or stage is None:
            self.test_dataset = self.dataset_cls(
                data_dir=self.data_dir,
                split="test",
                target_condition=self.target_condition,
            )

    def train_dataloader(self) -> DataLoader:
        if self.train_dataset is None:
            raise ValueError("Train dataset is not initialized.")
        return DataLoader(
            self.train_dataset,  # type: ignore
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            shuffle=True,
        )

    def val_dataloader(self) -> DataLoader:
        if self.val_dataset is None:
            raise ValueError("Validation dataset is not initialized.")
        return DataLoader(
            self.val_dataset,  # type: ignore
            batch_size=self.val_batch_size,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            shuffle=False,
        )

    def test_dataloader(self) -> DataLoader:
        if self.test_dataset is None:
            raise ValueError("Test dataset is not initialized.")
        return DataLoader(
            self.test_dataset,  # type: ignore
            batch_size=self.test_batch_size,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            shuffle=False,
        )



from setuptools import find_packages, setup

setup(
    name="chemeleon-dng",
    packages=find_packages(),
    install_requires=[
        "torch>=2.1.0",
        "pytorch-lightning>=2.1.0",
        "sacred",
        "ase",
        "torch-geometric",
        "torchmetrics",
        "ase",
        "tqdm",
        "wandb",
        "pydantic",
        "jupyterlab",
        "pymatgen",
        "fire",
    ],
)



import warnings
from pathlib import Path

import pandas as pd
from pymatgen.core import Lattice, Structure
from torch.utils.data import Dataset
from torch_geometric.data import Data

from chemeleon_dng.dataset.dataset_util import pmg_structure_to_pyg_data

warnings.filterwarnings("ignore", category=UserWarning, module="pymatgen")


class MPDataset(Dataset):
    """Dataset for Materials Project data (e.g., MP-20).
    """

    def __init__(
        self, data_dir: str | Path, split: str, target_condition: str | None = None
    ):
        self.data_dir = Path(data_dir)
        self.split = split
        self.target_condition = target_condition
        self.data = pd.read_csv(self.data_dir / f"{split}.csv")
        if target_condition is not None:
            self.condition = self.data[target_condition]
        else:
            self.condition = None

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, idx: int) -> Data:
        row = self.data.iloc[idx]

        # Get the condition for the target
        if self.condition is not None:
            cond = self.condition[idx]
        else:
            cond = None

        # Read CIF file
        str_cif = row["cif"]
        st = Structure.from_str(str_cif, fmt="cif")

        # Niggli reduction
        reduced_st = st.get_reduced_structure()
        canonical_st = Structure(
            lattice=Lattice.from_parameters(*reduced_st.lattice.parameters),
            species=reduced_st.species,
            coords=reduced_st.frac_coords,
            coords_are_cartesian=False,
        )

        # Convert PMG structure to PYG Data
        data = pmg_structure_to_pyg_data(canonical_st, cond=cond)
        return data



from abc import ABC, abstractmethod


class DiffusionModelBase(ABC):
    """Abstract base class for diffusion models.
    """

    @abstractmethod
    def q_sample(self):
        """Abstract method for the forward process (q_sample)."""
        raise NotImplementedError("q_sample method not implemented.")

    @abstractmethod
    def p_sample(self):
        """Abstract method for the reverse process (p_sample)."""
        raise NotImplementedError("p_sample method not implemented.")



from collections import OrderedDict

import torch
from ase import Atoms
from ase.build.tools import sort
from pydantic import BaseModel, ConfigDict, Field
from torch import Tensor


class CrystalBatch(BaseModel):
    """A schema for a batch of crystal structures.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    atom_types: Tensor
    lattices: Tensor
    frac_coords: Tensor
    num_atoms: Tensor
    batch: Tensor  # batch_idx
    num_graphs: int | None = None
    num_nodes: int | None = None
    noise_atom_types: Tensor | None = None
    noise_lattices: Tensor | None = None
    noise_frac_coords: Tensor | None = None


class Trajectory(BaseModel):
    """A schema for a trajectory of crystal structures.
    Each crystal structure is represented as a CrystalBatch.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")

    total_steps: int
    container: OrderedDict[int, CrystalBatch] = Field(default_factory=OrderedDict)

    def __getitem__(self, t: int):
        return self.container[t]

    def __setitem__(self, t: int, value: CrystalBatch) -> None:
        self.container[t] = value

    def __len__(self) -> int:
        return len(self.container)

    def get_atoms(self, t: int = 0, idx: int | None = None) -> Atoms | list[Atoms]:
        trajectory_step = self.container[t]

        # If atom type is greater than 100 + 1, set it to 0
        trajectory_step.atom_types = torch.where(
            trajectory_step.atom_types <= 101, trajectory_step.atom_types, 0
        )
        split_atom_types = torch.split(
            trajectory_step.atom_types, trajectory_step.num_atoms.tolist()
        )
        split_frac_coords = torch.split(
            trajectory_step.frac_coords, trajectory_step.num_atoms.tolist()
        )

        atoms_list = []
        for i, (frac_coords, atom_types) in enumerate(
            zip(split_frac_coords, split_atom_types, strict=False)
        ):
            atoms = Atoms(
                numbers=atom_types.detach().cpu().numpy(),
                cell=trajectory_step.lattices[i].detach().cpu().numpy(),
                pbc=True,
            )
            positions = frac_coords.detach().cpu().numpy()
            atoms.set_scaled_positions(positions)
            atoms_list.append(sort(atoms))
        if idx is None:
            return atoms_list
        else:
            return atoms_list[idx]

    def get_trajectory(self, idx: int | None = None):
        if idx is None:
            return [self.get_atoms(t, None) for t in range(self.total_steps + 1)]
        else:
            return [self.get_atoms(t, idx) for t in range(self.total_steps + 1)]



import copy
import enum

import torch
from torch import Tensor
from tqdm import tqdm

from chemeleon_dng.base_module import BaseModule
from chemeleon_dng.cspnet import CSPNet
from chemeleon_dng.diffusion.diffusion_scheduler import uniform_sample_t
from chemeleon_dng.diffusion.models import D3PM, DDPM, DSM
from chemeleon_dng.schema import CrystalBatch, Trajectory
from chemeleon_dng.script_util import create_model


class DiffusionModuleTask(enum.Enum):
    """Which task of the model.
    """

    CSP = enum.auto()  # Crystal Structure Prediction
    DNG = enum.auto()  # De Novo Generation
    GUIDE = enum.auto()  # Guiding Generation (Only available for DNG)


class DiffusionType(enum.Enum):
    """Type of diffusion model.
    """

    D3PM = enum.auto()  # Discrete Denoising Diffusion Probabilistic Models (Discrete)
    DDPM = enum.auto()  # Denoising Diffusion Probabilistic Model (Variance-preserving)
    DSM = enum.auto()  # Denising Score Matching (Variance-exploding)


class DiffusionModule(BaseModule):
    """Utilities for training and sampling diffusion models.
    """

    def __init__(
        self,
        *,
        task: DiffusionModuleTask,
        num_timesteps: int,
        diffusion_atom_type: D3PM,
        diffusion_lattice: DDPM,
        diffusion_frac_coord: DSM,
        model_configs: dict,
        optimizer_configs: dict,
    ):
        super().__init__(**optimizer_configs)
        self.save_hyperparameters()

        if task == DiffusionModuleTask.CSP:
            assert (
                diffusion_atom_type is None
            ), "Diffusion atom type should be None, when CSP"
        self.model = create_model(**model_configs, task=task.name)
        self.num_timesteps = num_timesteps
        self.diffusion_atom_type = diffusion_atom_type
        self.diffusion_lattice = diffusion_lattice
        self.diffusion_frac_coord = diffusion_frac_coord

    def q_sample(
        self,
        x_start: CrystalBatch,
        t: Tensor,
        noise_atom_types: Tensor | None = None,
        noise_lattices: Tensor | None = None,
        noise_frac_coords: Tensor | None = None,
    ) -> CrystalBatch:
        """Sample from the forward process q(x_t | x_0).
        """
        return CrystalBatch(
            atom_types=(
                self.diffusion_atom_type.q_sample(
                    x_start=x_start.atom_types,
                    t=t,
                    batch_idx=x_start.batch,
                    noise=noise_atom_types,
                )
                if self.diffusion_atom_type is not None
                else x_start.atom_types
            ),
            lattices=self.diffusion_lattice.q_sample(
                x_start=x_start.lattices,
                t=t,
                noise=noise_lattices,
            ),
            frac_coords=self.diffusion_frac_coord.q_sample(
                x_start=x_start.frac_coords,
                t=t,
                batch_idx=x_start.batch,
                noise=noise_frac_coords,
            ),
            num_atoms=x_start.num_atoms,
            batch=x_start.batch,
        )

    def calculate_loss(
        self,
        x_start: CrystalBatch,
        t: Tensor | None = None,
        cond_embeds: Tensor | None = None,
    ):
        if t is None:
            t = uniform_sample_t(
                num_timesteps=self.num_timesteps, batch_size=x_start.num_graphs
            ).to(self.device)
        # Create noise for atom types, lattices, and fractional coordinates
        if self.diffusion_atom_type is not None:
            noise_atom_types = torch.rand(
                len(x_start.atom_types), self.diffusion_atom_type.max_atoms
            ).to(x_start.atom_types.device)
        else:
            noise_atom_types = None
        noise_lattices = torch.randn_like(x_start.lattices)
        noise_frac_coords = torch.randn_like(x_start.frac_coords)
        # Sample from the forward process q(x_t | x_0)
        x_t = self.q_sample(
            x_start=x_start,
            t=t,
            noise_atom_types=noise_atom_types,
            noise_lattices=noise_lattices,
            noise_frac_coords=noise_frac_coords,
        )
        # Model prediction
        model_output = _model_prediction(
            model=self.model,
            x_t=x_t,
            t=t,
            cond_embeds=cond_embeds,
        )
        # Calculate the loss
        res = {}
        if self.diffusion_atom_type is not None:
            loss_atom_types = self.diffusion_atom_type.training_loss(
                pred_x_start_logits=model_output.atom_types,
                x_start=x_start.atom_types,
                x_t=x_t.atom_types,
                t=t,
                batch_idx=x_t.batch,
            )
        else:
            loss_atom_types = {}
        loss_lattices = self.diffusion_lattice.training_loss(
            model_output=model_output.lattices,
            noise=noise_lattices,
        )
        loss_frac_coords = self.diffusion_frac_coord.training_loss(
            model_output=model_output.frac_coords,
            noise=noise_frac_coords,
            t=t,
            batch_idx=x_t.batch,
        )
        res.update({f"{k}_atom_types": v for k, v in loss_atom_types.items()})
        res.update({f"{k}_lattices": v for k, v in loss_lattices.items()})
        res.update({f"{k}_frac_coords": v for k, v in loss_frac_coords.items()})
        total_loss = (
            (loss_atom_types["loss"] + loss_lattices["loss"] + loss_frac_coords["loss"])
            if self.diffusion_atom_type is not None
            else loss_lattices["loss"] + loss_frac_coords["loss"]
        )
        res.update({"total_loss": total_loss})
        return res

    def p_sample(
        self,
        x_t: CrystalBatch,
        t: Tensor,
        step_lr: float = 1e-5,
        cond_scale: float = 2.0,
        cond_embeds: Tensor | None = None,
        null_cond_embeds: Tensor | None = None,
    ) -> CrystalBatch:
        """Sample from the reverse process p(x_{t-1} | x_t).
        """
        # Predictor
        model_predictor_output = _model_prediction(
            model=self.model,
            x_t=x_t,
            t=t,
            cond_scale=cond_scale,
            cond_embeds=cond_embeds,
            null_cond_embeds=null_cond_embeds,
        )
        predictor_output = CrystalBatch(
            atom_types=(
                self.diffusion_atom_type.p_sample(
                    pred_x_start_logits=model_predictor_output.atom_types,
                    x_t=x_t.atom_types,
                    t=t,
                    batch_idx=x_t.batch,
                )
                if self.diffusion_atom_type is not None
                else x_t.atom_types
            ),
            lattices=self.diffusion_lattice.p_sample(
                model_output=model_predictor_output.lattices,
                x_t=x_t.lattices,
                t=t,
            ),
            frac_coords=self.diffusion_frac_coord.p_sample(
                model_output=model_predictor_output.frac_coords,
                x_t=x_t.frac_coords,
                t=t,
                batch_idx=x_t.batch,
                mode="predictor",
            ),
            num_atoms=x_t.num_atoms,
            batch=x_t.batch,
        )

        # Corrector (update only frac_coords)
        model_corrector_output = _model_prediction(
            model=self.model,
            x_t=predictor_output,
            t=t,
            cond_scale=cond_scale,
            cond_embeds=cond_embeds,
            null_cond_embeds=null_cond_embeds,
        )
        return CrystalBatch(
            atom_types=predictor_output.atom_types,
            lattices=predictor_output.lattices,
            frac_coords=self.diffusion_frac_coord.p_sample(
                model_output=model_corrector_output.frac_coords,
                x_t=predictor_output.frac_coords,
                t=t,
                batch_idx=x_t.batch,
                mode="corrector",
                step_lr=step_lr,
            ),
            num_atoms=x_t.num_atoms,
            batch=x_t.batch,
            num_graphs=x_t.num_graphs,
            num_nodes=x_t.num_nodes,
        )

    def sample(
        self,
        task: str,
        num_atoms: list[int] | Tensor,
        atom_types: list[int] | Tensor | None = None,
        return_trajectory: bool = False,
        step_lr: float = 1e-5,
        verbose: bool = True,
    ):
        if task.lower() == "csp":
            assert self.model.task.name.lower() == "csp", "Model task should be CSP"  # type: ignore
            if atom_types is not None:
                assert len(atom_types) == sum(
                    num_atoms
                ), "Length of atom_types and num_atoms should be the same."
            else:
                raise ValueError("atom_types must be provided for CSP task.")
        elif task.lower() == "dng":
            assert self.model.task.name.lower() == "dng", "Model task should be DNG"  # type: ignore
            if atom_types is not None:
                atom_types = None
                print("The given atom_type in arguments will be ignored for DNG task")
        elif task.lower() == "guide":
            assert self.model.task.name.lower() == "guide", "Model task should be GUIDE"  # type: ignore
        else:
            raise ValueError(
                f"Unknown task {task}. Supported tasks are 'csp', 'dng', and 'guide'."
            )
        assert num_atoms is not None

        # Construct an initial batch of crystal structures
        num_graphs = len(num_atoms)
        num_nodes = sum(num_atoms)
        x_t = CrystalBatch(
            atom_types=(
                torch.as_tensor(atom_types, dtype=torch.long, device=self.device)
                if atom_types is not None
                else torch.zeros(
                    num_nodes,
                    dtype=torch.long,
                    device=self.device,
                )
            ),
            lattices=torch.randn((num_graphs, 3, 3), device=self.device),
            frac_coords=torch.rand((num_nodes, 3), device=self.device) % 1.0,
            num_atoms=torch.as_tensor(num_atoms, device=self.device),
            batch=torch.tensor(
                [i for i, n in enumerate(num_atoms) for _ in range(n)],
                device=self.device,
            ),
            num_graphs=num_graphs,
            num_nodes=num_nodes,
        )

        # Create Trajectory object to store the generated structures
        trajectory = Trajectory(total_steps=self.num_timesteps)
        trajectory.container[self.num_timesteps] = x_t

        # Sample from the reverse process p(x_{t-1} | x_t)
        timesteps = range(self.num_timesteps, 0, -1)
        if verbose:
            timesteps = tqdm(timesteps, desc="Sampling")

        for timestep in timesteps:
            t = (
                torch.ones((len(num_atoms),), dtype=torch.long, device=self.device)
                * timestep
            )
            with torch.no_grad():
                x_t = self.p_sample(x_t=x_t, t=t, step_lr=step_lr)
            trajectory.container[timestep - 1] = x_t

        if return_trajectory:
            return trajectory
        return trajectory.get_atoms(t=0)


def _model_prediction(
    model: CSPNet,
    x_t: CrystalBatch,
    t: Tensor,
    cond_embeds: Tensor | None = None,
    null_cond_embeds: Tensor | None = None,
    cond_scale: float = 1.0,
) -> CrystalBatch:
    if cond_embeds is not None:
        assert null_cond_embeds is not None
        cond_model_output = model(
            atom_types=x_t.atom_types,
            lattices=x_t.lattices,
            frac_coords=x_t.frac_coords,
            num_atoms=x_t.num_atoms,
            batch_idx=x_t.batch,
            t=t,
            cond_embeds=cond_embeds,
        )
        null_model_output = model(
            atom_types=x_t.atom_types,
            lattices=x_t.lattices,
            frac_coords=x_t.frac_coords,
            num_atoms=x_t.num_atoms,
            batch_idx=x_t.batch,
            t=t,
            cond_embeds=null_cond_embeds,
        )
        output = copy.deepcopy(cond_model_output)
        output.atom_types = (
            1 - cond_scale
        ) * null_model_output.atom_types + cond_scale * cond_model_output.atom_types
        output.lattices = (
            1 - cond_scale
        ) * null_model_output.lattices + cond_scale * cond_model_output.lattices
        output.frac_coords = (
            (1 - cond_scale) * null_model_output.frac_coords
            + cond_scale * cond_model_output.frac_coords
        )
    else:
        output = model(
            atom_types=x_t.atom_types,
            lattices=x_t.lattices,
            frac_coords=x_t.frac_coords,
            num_atoms=x_t.num_atoms,
            batch_idx=x_t.batch,
            t=t,
        )
    return output



import torch
from ase import Atoms
from pymatgen.core import Lattice, Structure
from torch_geometric.data import Batch, Data


def pmg_structure_to_pyg_data(pmg_structure: Structure, **kwargs) -> Data:
    return Data(
        pos=torch.tensor(pmg_structure.frac_coords, dtype=torch.float),
        atom_types=torch.tensor(pmg_structure.atomic_numbers, dtype=torch.long),
        frac_coords=torch.tensor(pmg_structure.frac_coords, dtype=torch.float),
        cart_coords=torch.tensor(pmg_structure.cart_coords, dtype=torch.float),
        lattices=torch.tensor(
            pmg_structure.lattice.matrix, dtype=torch.float
        ).unsqueeze(0),
        num_atoms=torch.tensor([len(pmg_structure)]),
        **kwargs,
    )


def batch_to_atoms_list(batch: Batch, frac_coords: bool = True) -> list[Atoms]:
    atoms_list = []
    for data in batch.to_data_list():
        atoms = Atoms(
            numbers=data.atom_types.detach().cpu().numpy(),
            cell=data.lattices.squeeze(0).detach().cpu().numpy(),
            pbc=True,
        )
        if frac_coords:
            positions = data.frac_coords.detach().cpu().numpy()
            atoms.set_scaled_positions(positions)
        else:
            positions = data.cart_coords.detach().cpu().numpy()
            atoms.set_positions(positions)
        atoms_list.append(atoms)
    return atoms_list


def batch_to_structure_list(batch: Batch, frac_coords: bool = True) -> list[Structure]:
    structure_list = []
    for data in batch.to_data_list():
        # Get atomic numbers (convert to list for pymatgen Structure)
        atomic_numbers = data.atom_types.detach().cpu().numpy().tolist()

        # Create lattice from lattice matrix
        lattice = Lattice(data.lattices.squeeze(0).detach().cpu().numpy())

        if frac_coords:
            coords = data.frac_coords.detach().cpu().numpy()
            structure = Structure(
                lattice=Lattice.from_parameters(*lattice.parameters),
                species=atomic_numbers,
                coords=coords,
                coords_are_cartesian=False,
            )
        else:
            coords = data.cart_coords.detach().cpu().numpy()
            structure = Structure(
                lattice=Lattice.from_parameters(*lattice.parameters),
                species=atomic_numbers,
                coords=coords,
                coords_are_cartesian=True,
            )
        structure_list.append(structure)
    return structure_list
